import numpy as np
import torch
from torch.utils.data import DataLoader
import sys
from tqdm.notebook import tqdm
import warnings

from . import models
from . import ksg

class EarlyStopper:
    """
    Early stopping that returns best weights
    trying to replicate the Keras callback
    """
    def __init__(self, patience=1):
        self.patience = patience
        self.counter = 0
        self.min_validation_loss = float('inf')
        self.best_state = None

    def early_stop(self, validation_loss, model):
        if validation_loss < self.min_validation_loss:
            self.min_validation_loss = validation_loss
            self.counter = 0
            self.best_state = model.state_dict()
        else:
            self.counter += 1
            if self.counter >= self.patience:
                return self.best_state
        return False


def train(model, X_train, Y_train, X_test, Y_test,
          batch_size=512, lr=0.0001, epochs=300, patience=30, 
          quiet=True):
    """
    training loop for LMI models

    :param model: LMI model

    :param X_train: train samples, shape (N_samples, N_features)
    :param Y_train: train samples, shape (N_samples, N_features)
    :param X_test: test samples, shape (N_samples, N_features)
    :param Y_test: test samples, shape (N_samples, N_features)

    :param batch_size: samples per batch, defaults to 512
    :param lr: learning rate for Adam optimizer, defaults to 1e-4
    :param epochs: max number of epochs, defaults to 300
    :param patience: epochs without val. loss decline before early stopping, 
                     defaults to 300
    :param quiet: suppress training progress display, defaults to True
    """
    
    
    optimizer = torch.optim.Adam(model.parameters(), lr=lr, eps=1e-07)
    
    train_dataloader = DataLoader(list(zip(X_train, Y_train)), 
                                batch_size=batch_size, 
                                shuffle=True)

    val_dataloader = DataLoader(list(zip(X_test, Y_test)), 
                                batch_size=batch_size, 
                                shuffle=True)
    
    val_losses = []

    early_stopper = EarlyStopper(patience=patience)
    
    # with tqdm(range(epochs), unit='Epoch', disable=quiet) as tepoch:
    for epoch in range(epochs):
        
        for i, (X, Y) in enumerate(train_dataloader):
            
            model.train() 
            model_loss = model.learning_loss(X, Y)

            optimizer.zero_grad()
            model_loss.backward()
            optimizer.step()
        
        # validation 
        with torch.no_grad():
            epoch_validate_loss = []
            for i, data in enumerate(val_dataloader):
                X, Y = data
                epoch_validate_loss.append(model.learning_loss(X, Y).item())
            val_losses.append(np.mean(epoch_validate_loss))

        # tepoch.set_postfix(val_loss=val_losses[-1])
        # early stopping
        es = early_stopper.early_stop(val_losses[-1], model)
        if es or epoch==epochs:
            if not quiet:
                sys.stdout.write('epoch %d (of max %d) %s \U0001F389\U0001F389\r' 
                         %(epoch, epochs, '\U0001F33B'*int(10*(epoch/epochs))))
                sys.stdout.flush()
                sys.stdout.write('\n')
                print("success! training stopped at epoch %d" % epoch)
                print('final validation loss:', val_losses[-1])
            model.load_state_dict(es)
            return
        
        sys.stdout.write('epoch %d (of max %d) %s\r' 
                         %(epoch, epochs, '\U0001F33B'*int(10*(epoch/epochs))))
        sys.stdout.flush()


def learn_representation(Xs, Ys, train_indices, test_indices,
       regularizer='models.AECross', 
       alpha=1, lam=1,
       N_dims=8, batch_size=512, lr=0.0001, epochs=300,
       validation_split=0.3, patience=30, quiet=True, device='cpu'):
    """
    train paired AE model and embed data

    :param Xs:
    :param Ys:

    :param train_indices:
    :param test_indices:

    :param regularizer:
    :param alpha:
    :param lam:
    :param N_dims:
    
    :param batch_size:
    :param lr:
    :param epochs:
    :param validation_split:
    :param patience:
    :param quiet:
    """
    X_train = Xs[train_indices]
    Y_train = Ys[train_indices]
    
    X_test = Xs[test_indices]
    Y_test = Ys[test_indices]
    
    if 128 < N_dims or 128 < N_dims:
        warnings.warn("Hidden layer smaller than latent dimension. Consider reducing N_dims")
        
    # assert X_train.shape[1] // 4 > 0, "Hidden layer with size 0. Consider tiling input."
    # assert Y_train.shape[1] // 4 > 0, "Hidden layer with size 0. Consider tiling input."
    
    model = eval(regularizer)(X_train.shape[1], Y_train.shape[1], N_dims, 
                              alpha=alpha, lam=lam).to(device)
    
    train(model, X_train, Y_train, X_test, Y_test, 
          batch_size=batch_size, lr=lr, epochs=epochs, patience=patience,
          quiet=quiet)
    
    with torch.no_grad():
        model.eval()
        Zx, Zy = model.encode(Xs, Ys)
        
        Zx, Zy = Zx.cpu(), Zy.cpu()


        return Zx.cpu(), Zy.cpu(), model

def estimate(Xs, Ys, regularizer='models.AECross', 
         alpha=1, lam=1,
         N_dims=8, validation_split=0.5, estimate_on_val=True,
         batch_size=512, lr=0.0001, epochs=300, patience=30,
         quiet=True, device=None):
    """
    return pMIs, with NaNs for points not included in KSG estimate

    :param Xs:
    :param Ys:

    :param regularizer:
    :param alpha:
    :param lam:
    :param N_dims:
    
    :param batch_size:
    :param lr:
    :param epochs:
    :param validation_split:
    :param patience:
    :param quiet:
    """

    if device == None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'


    Xs = torch.from_numpy(np.nan_to_num((Xs - Xs.mean(axis=0)) / Xs.std(axis=0))).float().to(device)
    Ys = torch.from_numpy(np.nan_to_num((Ys - Ys.mean(axis=0)) / Ys.std(axis=0))).float().to(device)

    Xs = torch.clip(Xs, min=-10, max=10)
    Ys = torch.clip(Ys, min=-10, max=10)

    assert len(Xs) == len(Ys), "X and Y must be same length!"
    
    N_train = int(len(Xs) * (1 - validation_split))
    
    indices = np.arange(len(Xs))
    np.random.shuffle(indices)
    
    train_indices = indices[:N_train]
    test_indices = indices[N_train:]

    
    Zx, Zy, model = learn_representation(Xs, Ys, train_indices, test_indices,
                regularizer=regularizer, N_dims=N_dims, batch_size=batch_size,
                patience=patience, epochs=epochs, 
                lr=lr, quiet=quiet,
                alpha=alpha, lam=lam, device=device)

    if torch.isnan(Zx).any() or torch.isnan(Zy).any():
        warnings.warn("NaNs in embedding! converted to 0s")

    Zx = torch.nan_to_num(Zx)
    Zy = torch.nan_to_num(Zy)
    
    estimate = 0

    if estimate_on_val:

        # make nan array
        estimate = np.zeros(len(Xs))
        estimate += np.NaN

        # fill val pMIs
        estimate[indices[N_train:]] = ksg.mi(Zx.cpu()[indices[N_train:]], 
        Zy.cpu()[indices[N_train:]])
    
    else:
        estimate += ksg.mi(Zx.cpu(), Zy.cpu())
    
    return estimate, (Zx.cpu(), Zy.cpu()), model



import numpy as np
import torch 
import torch.nn as nn

class AECross(nn.Module):
    
    def __init__(self, x_dim, y_dim, latent_size, alpha=1, lam=1):
        """

        Paired AE models with cross predictive regularization
        
        """
        
        super(AECross, self).__init__()
        
        # choosing hidden layer sizes
        # Lx = int(2**np.floor(np.log2(x_dim)))
        # Ly = int(2**np.floor(np.log2(y_dim)))
        
        Lx, Ly = 1024, 1024
        
        if x_dim > 2048:
            Lx = int(2**np.floor(np.log2(x_dim)))
        if y_dim > 2048:
            Ly = int(2**np.floor(np.log2(y_dim)))

        self.x_encoder = nn.Sequential(nn.Linear(x_dim, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.y_encoder = nn.Sequential(nn.Linear(y_dim, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.xx_decoder = nn.Sequential(nn.Linear(latent_size, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, x_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.yy_decoder = nn.Sequential(nn.Linear(latent_size, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, y_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.yx_decoder = nn.Sequential(nn.Dropout(p=0.5),
                                        nn.Linear(latent_size, Lx//4),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Lx//4, Lx//2),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Lx//2, Lx),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Lx, x_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.xy_decoder = nn.Sequential(nn.Dropout(p=0.5),
                                        nn.Linear(latent_size, Ly//4),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Ly//4, Ly//2),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Ly//2, Ly),
                                        nn.LeakyReLU(negative_slope=0.2),
                                        nn.Dropout(p=0.5),
                                        nn.Linear(Ly, y_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))

        self.alpha = alpha
        self.lam = lam

        def init_weights(m):
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight)
                m.bias.data.fill_(0.0)

        for net in [self.x_encoder, self.y_encoder, self.xx_decoder, self.yy_decoder, 
                    self.xy_decoder, self.yx_decoder]:
            net.apply(init_weights)
        
    def encode(self, x_samples, y_samples):
        Zx = self.x_encoder(x_samples)
        Zy = self.y_encoder(y_samples)
        return Zx, Zy

    def cross_decode(self, Zx, Zy):
        y_hat = self.xy_decoder(Zx)
        x_hat = self.yx_decoder(Zy)
        return x_hat, y_hat
    
    def decode(self, Zx, Zy):
        x_hat = self.xx_decoder(Zx)
        y_hat = self.yy_decoder(Zy)
        return x_hat, y_hat
    
    def rec_loss(self, hat, samples):
        return torch.nn.functional.mse_loss(hat, samples, reduction='mean')

    def learning_loss(self, x_samples, y_samples):
        
        Zx, Zy = self.encode(x_samples, y_samples)
        Xh, Yh = self.decode(Zx, Zy)
        cXh, cYh = self.cross_decode(Zx, Zy)
        
        auto_loss = self.rec_loss(Xh, x_samples) + self.rec_loss(Yh, y_samples)
        cross_loss = self.rec_loss(cXh, x_samples) + self.rec_loss(cYh, y_samples)
        # print(auto_loss, cross_loss)
        
        return self.lam*cross_loss + self.alpha*auto_loss


class AEMINE(nn.Module):
    
    def __init__(self, x_dim, y_dim, latent_size, alpha=1, lam=1):
        """

        Paired AE models with MINE regularization

        """
        
        super(AEMINE, self).__init__()
        
        # choosing hidden layer sizes
        # Lx = int(2**np.floor(np.log2(x_dim)))
        # Ly = int(2**np.floor(np.log2(y_dim)))
        Lx, Ly = 1024, 1024
        
        self.x_encoder = nn.Sequential(nn.Linear(x_dim, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.y_encoder = nn.Sequential(nn.Linear(y_dim, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.xx_decoder = nn.Sequential(nn.Linear(latent_size, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, x_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.yy_decoder = nn.Sequential(nn.Linear(latent_size, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, y_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.T_func = nn.Sequential(nn.Linear(latent_size*2, latent_size),
                                    nn.ReLU(),
                                    nn.Linear(latent_size, 1))

        self.alpha = alpha
        self.lam = lam

        def init_weights(m):
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight)
                m.bias.data.fill_(0.0)

        for net in [self.x_encoder, self.y_encoder, self.xx_decoder, self.yy_decoder, 
                    self.T_func]:
            net.apply(init_weights)
        
    def encode(self, x_samples, y_samples):
        Zx = self.x_encoder(x_samples)
        Zy = self.y_encoder(y_samples)
        return Zx, Zy

    def MINELoss(self, x_samples, y_samples):  
        
        sample_size = y_samples.shape[0]
        random_index = torch.randint(sample_size, (sample_size,)).long()

        y_shuffle = y_samples[random_index]

        T0 = self.T_func(torch.cat([x_samples,y_samples], dim = -1))
        T1 = self.T_func(torch.cat([x_samples,y_shuffle], dim = -1))

        lower_bound = T0.mean() - torch.log(T1.exp().mean())
        return -lower_bound
    
    def decode(self, Zx, Zy):
        x_hat = self.xx_decoder(Zx)
        y_hat = self.yy_decoder(Zy)
        return x_hat, y_hat
    
    def rec_loss(self, hat, samples):
        return torch.nn.functional.mse_loss(hat, samples, reduction='mean')

    def learning_loss(self, x_samples, y_samples):
        
        Zx, Zy = self.encode(x_samples, y_samples)
        Xh, Yh = self.decode(Zx, Zy)
        
        auto_loss = self.rec_loss(Xh, x_samples) + self.rec_loss(Yh, y_samples)
        cross_loss = self.MINELoss(Zx, Zy)
        
        return self.lam*cross_loss + self.alpha*auto_loss


class AEInfoNCE(nn.Module):
    
    def __init__(self, x_dim, y_dim, latent_size, alpha=1, lam=1):
        """

        Paired AE models with InfoNCE regularization
        
        """
        
        super(AEInfoNCE, self).__init__()
        
        # choosing hidden layer sizes
        # Lx = int(2**np.floor(np.log2(x_dim)))
        # Ly = int(2**np.floor(np.log2(y_dim)))
        Lx, Ly = 1024, 1024
        
        self.x_encoder = nn.Sequential(nn.Linear(x_dim, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.y_encoder = nn.Sequential(nn.Linear(y_dim, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, latent_size),
                                      nn.LeakyReLU(negative_slope=0.2))
        
        self.xx_decoder = nn.Sequential(nn.Linear(latent_size, Lx//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//4, Lx//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx//2, Lx),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Lx, x_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.yy_decoder = nn.Sequential(nn.Linear(latent_size, Ly//4),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//4, Ly//2),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly//2, Ly),
                                      nn.LeakyReLU(negative_slope=0.2),
                                      nn.Linear(Ly, y_dim),)
                                    #   nn.LeakyReLU(negative_slope=0.2))
        
        self.F_func = nn.Sequential(nn.Linear(latent_size+latent_size, latent_size),
                                    nn.ReLU(),
                                    nn.Linear(latent_size, 1),
                                    nn.Softplus())

        self.alpha = alpha
        self.lam = lam

        def init_weights(m):
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight)
                m.bias.data.fill_(0.0)

        for net in [self.x_encoder, self.y_encoder, self.xx_decoder, self.yy_decoder, 
                    self.F_func]:
            net.apply(init_weights)
        
    def encode(self, x_samples, y_samples):
        Zx = self.x_encoder(x_samples)
        Zy = self.y_encoder(y_samples)
        return Zx, Zy

    def InfoNCELoss(self, x_samples, y_samples):  
        
        sample_size = y_samples.shape[0]

        x_tile = x_samples.unsqueeze(0).repeat((sample_size, 1, 1))
        y_tile = y_samples.unsqueeze(1).repeat((1, sample_size, 1))

        T0 = self.F_func(torch.cat([x_samples,y_samples], dim = -1))
        T1 = self.F_func(torch.cat([x_tile, y_tile], dim = -1))  #[sample_size, sample_size, 1]

        lower_bound = T0.mean() - (T1.logsumexp(dim = 1).mean() - np.log(sample_size)) 
        return -lower_bound
    
    def decode(self, Zx, Zy):
        x_hat = self.xx_decoder(Zx)
        y_hat = self.yy_decoder(Zy)
        return x_hat, y_hat
    
    def rec_loss(self, hat, samples):
        return torch.nn.functional.mse_loss(hat, samples, reduction='mean')

    def learning_loss(self, x_samples, y_samples):
        
        Zx, Zy = self.encode(x_samples, y_samples)
        Xh, Yh = self.decode(Zx, Zy)
        
        auto_loss = self.rec_loss(Xh, x_samples) + self.rec_loss(Yh, y_samples)
        cross_loss = self.InfoNCELoss(Zx, Zy)
        
        return self.lam*cross_loss + self.alpha*auto_loss



from scipy.special import digamma
from sklearn.neighbors import BallTree, KDTree
import numpy as np
import numpy.linalg as la

############# adapted from NPEET ############
### https://github.com/gregversteeg/NPEET ###
#############################################

def add_noise(x, intens=1e-10):
    # small noise to break degeneracy
    return x + intens * np.random.random_sample(x.shape)


def query_neighbors(tree, x, k):
    return tree.query(x, k=k + 1)[0][:, k]


def count_neighbors(tree, x, r):
    return tree.query_radius(x, r, count_only=True)

def avgdigamma(points, dvec):
    """
    tweaked this to return list of psi(nx) for pointwise estimate
    """
    tree = build_tree(points)
    dvec = dvec - 1e-15
    num_points = count_neighbors(tree, points, dvec)
    return digamma(num_points)


def build_tree(points):
    if points.shape[1] >= 20:
        return BallTree(points, metric="chebyshev")
    return KDTree(points, metric="chebyshev")

def lnc_correction(tree, points, k, alpha):
    """
    in my experience this doesn't work in practice for N_dims > 1
    """
    e = 0
    n_sample = points.shape[0]
    for point in points:
        # Find k-nearest neighbors in joint space, p=inf means max norm
        knn = tree.query(point[None, :], k=k + 1, return_distance=False)[0]
        knn_points = points[knn]
        # Substract mean of k-nearest neighbor points
        knn_points = knn_points - knn_points[0]
        # Calculate covariance matrix of k-nearest neighbor points, obtain eigen vectors
        covr = knn_points.T @ knn_points / k
        _, v = la.eig(covr)
        # Calculate PCA-bounding box using eigen vectors
        V_rect = np.log(np.abs(knn_points @ v).max(axis=0)).sum()
        # Calculate the volume of original box
        log_knn_dist = np.log(np.abs(knn_points).max(axis=0)).sum()

        # Perform local non-uniformity checking and update correction term
        if V_rect < log_knn_dist + np.log(alpha):
            print(V_rect)
            print(log_knn_dist + np.log(alpha))
            print()
            e += (log_knn_dist - V_rect) / n_sample
    return e

def entropy(x, k=3, base=2):
    """The classic K-L k-nearest neighbor continuous entropy estimator
    x should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
    if x is a one-dimensional scalar and we have four samples
    """
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    x = np.asarray(x)
    n_elements, n_features = x.shape
    x = add_noise(x)
    tree = build_tree(x)
    nn = query_neighbors(tree, x, k)
    const = digamma(n_elements) - digamma(k) + n_features * np.log(2)
    return (const + n_features * np.log(nn).mean()) / np.log(base)

def mi(x, y, z=None, k=3, base=2, alpha=0):
    """Mutual information of x and y (conditioned on z if z is not None)
    x, y should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
    if x is a one-dimensional scalar and we have four samples
    """
    
    assert len(x) == len(y), "Arrays should have same length"
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    
    x, y = np.asarray(x), np.asarray(y)
    x, y = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
    
    x = add_noise(x)
    y = add_noise(y)
    points = [x, y]
    
    if z is not None:
        z = np.asarray(z)
        z = z.reshape(z.shape[0], -1)
        points.append(z)
    points = np.hstack(points)
    # Find nearest neighbors in joint space, p=inf means max-norm
    tree = build_tree(points)
    dvec = query_neighbors(tree, points, k)
    if z is None:
        a, b, c, d = (
            avgdigamma(x, dvec),
            avgdigamma(y, dvec),
            digamma(k),
            digamma(len(x)),
        )
        if alpha > 0:
            d += lnc_correction(tree, points, k, alpha)
    else:
        xz = np.c_[x, z]
        yz = np.c_[y, z]
        a, b, c, d = (
            avgdigamma(xz, dvec),
            avgdigamma(yz, dvec),
            avgdigamma(z, dvec),
            digamma(k),
        )

#     mi_estimate = c + d - np.mean(a) - np.mean(b)
#     nxys = a + b
#     return mi_estimate, nxys
    return (c + d - a - b) / np.log(base)

# DISCRETE ESTIMATORS
def entropyd(sx, base=2):
    """Discrete entropy estimator
    sx is a list of samples
    """
    unique, count = np.unique(sx, return_counts=True, axis=0)
    # Convert to float as otherwise integer division results in all 0 for proba.
    proba = count.astype(float) / len(sx)
    # Avoid 0 division; remove probabilities == 0.0 (removing them does not change the entropy estimate as 0 * log(1/0) = 0.
    proba = proba[proba > 0.0]
    return np.sum(proba * np.log(1.0 / proba)) / np.log(base)

def centropyd(x, y, base=2):
    """The classic K-L k-nearest neighbor continuous entropy estimator for the
    entropy of X conditioned on Y.
    """
    xy = np.c_[x, y]
    return entropyd(xy, base) - entropyd(y, base)


def midd(x, y, base=2):
    """Discrete mutual information estimator
    Given a list of samples which can be any hashable object
    """
    assert len(x) == len(y), "Arrays should have same length"
    return entropyd(x, base) - centropyd(x, y, base)


def out_of_sample_avgdigamma(obs_points, est_points, dvec):
    """
    here we compute psi(nx) for query points that are not
    necessarily part of observed samples

    :param obs_points:
    :param est_points:
    :param dvec:
    """

    tree = build_tree(obs_points)
    dvec = dvec - 1e-15
    num_points = count_neighbors(tree, est_points, dvec)
    return digamma(num_points)

def out_of_sample_pmi(obs_x, obs_y, est_x, est_y, k=3, base=2):
    """
    
    
    """
    obs_x = add_noise(obs_x)
    obs_y = add_noise(obs_y)
    
    est_x = add_noise(est_x)
    est_y = add_noise(est_y)
    
    obs_points = np.hstack([obs_x, obs_y])
    est_points = np.hstack([est_x, est_y])
    tree = build_tree(obs_points)
    dvec = query_neighbors(tree, est_points, k)
    
    a = out_of_sample_avgdigamma(obs_x, est_x, dvec)
    b = out_of_sample_avgdigamma(obs_y, est_y, dvec)
    c = digamma(k)
    d = digamma(len(obs_x))
    
    return (c+d-a-b) / np.log(base)
